"""A script containing information about handing a project file read
from the metadata.
"""

import os
from typing import TypeVar, List, Callable, Tuple, TypeAlias
from abc import ABC, abstractmethod
from jammies.log import Logger
from jammies.utils import get_default, get_or_default
from jammies.struct.codec import DictObject, DictCodec

PostProcessor: TypeAlias = Callable[[Logger, str], bool]
"""A method which takes in the directory to execute within and returns
a boolean represent whether the execution was successful.
"""

class ProjectFile(ABC):
    """An abstract class containing information about a file associated with the project.
    """

    @abstractmethod
    def __init__(self, codec: 'ProjectFileCodec',
            name: str = "", rel_dir: str = os.curdir,
            post_processor: Tuple[PostProcessor, DictObject] | None = None,
            extra: DictObject | None = None) -> None:
        """
        Parameters
        ----------
        name : str (default '')
            The name of the project file.
        rel_dir : str (default '.')
            The directory the project file is located.
        post_processor : `PostProcessor`, `DictObject` (default `None`)
            The post processor to apply to the project file with any additional metadata.
        extra : dict[str, Any] (default `None`)
            Extra data defined by the user.
        """
        super().__init__()
        self._codec: 'ProjectFileCodec' = codec
        self.name: str = name
        self.dir: str = rel_dir
        self.post_processor: Tuple[PostProcessor, DictObject] = post_processor
        self.extra: DictObject = {} if extra is None else extra

    @abstractmethod
    def registry_name(self) -> str:
        """Returns the registered name of the project file.

        Returns
        -------
        str
            The registered name of the project file.
        """

    def codec(self) -> 'ProjectFileCodec':
        """Returns the codec used to encode and decode this project file.

        Returns
        -------
        ProjectFileCodec
            The codec used to encode and decode this project file.
        """
        return self._codec

    @abstractmethod
    def setup(self, root_dir: str, ignore_sub_directory: bool = False) -> bool:
        """Sets up the project file for usage.

        Parameters
        ----------
        root_dir : str
            The root directory to set up the project file in.
        ignore_sub_directory : bool (default `False`)
            When `False`, uses the root directory appended with the file's subdirectory.
        """
        # Create base if subdirectory should exist
        if not ignore_sub_directory:
            os.makedirs(self.create_path(root_dir), exist_ok = True)
        return False

    def create_path(self, root_dir: str, *paths: str) -> str:
        """Constructs a path from the root directory through the relative
        directory and any additional paths specified.

        Parameters
        ----------
        root_dir : str
            The root directory to create the path from.
        *paths : str
            The paths after the project file's relative directory.

        Returns
        -------
        str
            The newly created path.
        """
        fpath: List[str] = [root_dir, self.dir]
        fpath += paths
        return os.sep.join(fpath)

PF = TypeVar('PF', bound = ProjectFile)
"""The type of the project file."""

class ProjectFileCodec(DictCodec[PF]):
    """An abstract, generic encoder and decoder between a dictionary and a ProjectFile.

    Types
    -----
    PF
        The type of the project file to be encoded or decoded to.
    """

    def __init__(self, registrar) -> None:
        """Initializes a codec for a `ProjectFile`.

        Parameters
        ----------
        registrar : `JammiesRegistrar`
            The registrar used to register the components for the project.
        """
        self.registrar = registrar

    def decode(self, obj: DictObject) -> PF:
        return self.decode_type(obj,
            name = get_or_default(obj, 'name', ProjectFile),
            rel_dir = get_or_default(obj, 'dir', ProjectFile, param = 'rel_dir'),
            post_processor = self.__decode_post_processor(
                get_or_default(obj, 'post_processor', ProjectFile)
            ),
            extra = get_or_default(obj, 'extra', ProjectFile)
        )

    def __decode_post_processor(self,
            post_processor: DictObject | None) -> Tuple[PostProcessor, DictObject]:
        """Decodes a post processor from its object.

        Parameters
        ----------
        post_processor : `DictObject`
            The encoded post processor.
        
        Returns
        -------
        (`PostProcessor`, `DictObject`)
            The post processor and its additional parameters.
        """
        return (self.registrar.get_post_processor(post_processor['type']), post_processor) \
            if post_processor is not None else None

    @abstractmethod
    def decode_type(self, obj: DictObject, **kwargs: DictObject) -> PF:
        """Decodes a dictionary to the specific ProjectFile type.

        Parameters
        ----------
        obj : Dict[str, Any]
            The dictionary containing the data for the ProjectFile.

        Returns
        -------
        PF
            The decoded ProjectFile.
        """

    def encode(self, obj: PF) -> DictObject:
        dict_obj: DictObject = {}
        dict_obj['type'] = obj.registry_name()
        if obj.name:
            dict_obj['name'] = obj.name
        if obj.dir != get_default(ProjectFile, 'rel_dir'):
            dict_obj['dir'] = obj.dir
        if obj.post_processor:
            dict_obj['post_processor'] = obj.post_processor[1]
        if obj.extra:
            dict_obj['extra'] = obj.extra

        return self.encode_type(obj, dict_obj)

    @abstractmethod
    def encode_type(self, obj: PF, dict_obj: DictObject) -> DictObject:
        """Encodes a specific ProjectFile type to the dictionary.

        Parameters
        ----------
        obj : PF
            The ProjectFile containing the data.
        dict_obj : Dict[str, Any]
            The dictionary containing some common encoded data.

        Returns
        -------
        Dict[str, Any]
            The encoded ProjectFile in a dictionary.
        """



"""A script containing helper methods used throughout the package.
"""

import os
import re
import inspect
from typing import Any, Dict, Callable, TypeVar, Tuple
from io import IOBase, BytesIO
from mimetypes import guess_extension
from zipfile import ZipFile
from urllib.parse import urlparse
import requests

def get_default(func: Callable[..., Any], param: str) -> Any | None:
    """Gets the default value of a function parameter, or `None` if not applicable.
    
    Parameters
    ----------
    func : Callable[..., Any]
        The function to check.
    param : str
        The name of the parameter.

    Returns
    -------
    Any | None
        The default value of the parameter, or `None` if not applicable.
    """
    param_sig: inspect.Parameter = inspect.signature(func).parameters[param]
    return None if param_sig.default is inspect.Parameter.empty else param_sig.default

def get_or_default(dict_obj: Dict[str, Any], key: str, func: Callable[..., Any],
        param: str | None = None) -> Any | None:
    """Gets the value within a dictionary, or the default from the function if none is specified.

    Parameters
    ----------
    dict_obj : Dict[str, Any]
        The dictionary containing the data of an object.
    key : str
        The key to obtain the value of the dictionary from.
    func : Callable[..., Any]
        The function to check the default of if the key does not exist in the dictionary.
    param : str | None (default None)
        The name of the parameter containing the default value. When `None`, defaults to the `key`.
    
    Returns
    -------
    Any | None
        The value of the key, the default value, or `None`.
    """
    if param is None:
        param: str = key

    return dict_obj[key] if key in dict_obj else get_default(func, param)

I = TypeVar('I')
"""The type of the input."""

def input_with_default(func: Callable[..., Any], param: str, text: str) -> I:
    """Read a string from standard input and defaults if no value is specified.
    
    Parameters
    ----------
    func : Callable[..., Any]
        The function to check.
    param : str
        The name of the parameter.
    text : str
        The text to display for input.

    Returns
    -------
    Any
        The input, or the default when not specified.
    """
    def_val: I = get_default(func, param)
    val: I = input(f"{text} (default '{def_val}'): ")
    return val if val else def_val

def input_yn_default(text: str, yes_or_no: bool) -> bool:
    """Accepts an input of a yes / no answer which defaults to the
    specified value.

    Parameters
    ----------
    text: str
        The text to display for input.
    yes_or_no: bool
        When `True`, the input will default to 'yes'.
    
    Returns
    -------
    bool
        `True` if the input said 'yes'.
    """
    def cap_when_true(text: str, cap: bool) -> str:
        return text.capitalize() if cap else text

    input_string: str = f"{text} ({cap_when_true('y', yes_or_no)}" \
        + f"/{cap_when_true('n', not yes_or_no)})? "

    while True:
        if yn_input := input(input_string):
            if (yes_no := yn_input.lower()[0]) in ['y', 'n']:
                return yes_no != 'n'
            print('Answer provided was not y/n, please input y/n.')
            continue
        return yes_or_no

def unzip(file: str | IOBase, out_dir: str = os.curdir) -> None:
    """Unzips the file or stream to the specified directory.

    Parameters
    ----------
    file : str | io.IOBase
        The file or stream of the zip file.
    out_dir : str (default '.')
        The directory to extract the zip file to.
    """
    with ZipFile(file, 'r') as zip_ref: # type: ZipFile
        zip_ref.extractall(out_dir)

_FILENAME_PARAM_REGEX: str = \
    r'filename=(?:([A-Za-z0-9\!\#\$\%\&\'\*\+\-\.\^\_\`\|\~]+)|(?:\"([^\"]*)\"))'
"""Regex for getting the filename from the content-disposition header using the
[RFC6266](https://datatracker.ietf.org/doc/html/rfc6266#section-4.1) spec.
"""

_FILENAME_STAR_PARAM_REGEX: str = \
    r'filename\*=(?:[A-Za-z0-9\!\#\$\%\&\+\-\^\_\`\{\}\~]+)\'[A-Za-z0-9\%\-]*\'' \
        + r'((?:%[0-9A-Fa-f]{2}|[A-Za-z0-9\!\#\$\%\&\+\-\.\^\_\`\|\~])*)'
"""Regex for getting the filename* from the content-disposition header using the
[RFC8187](https://datatracker.ietf.org/doc/html/rfc8187) spec.
"""

_CONTENT_DISPOSITION: str = 'content-disposition'
"""The header for the content disposition."""

_CONTENT_TYPE: str = 'content-type'
"""The header for the content type."""

def download_file(url: str, handler: Callable[[requests.Response, str], bool],
        stream: bool = True) -> bool:
    """Downloads a file from the specified url via a GET request and handles the response
    bytes as specified.
    
    Parameters
    ----------
    url : str
        The url to download the file from.
    handler : (requests.Response, str) -> bool
        A function which takes in the response and filename and returns whether the file was
        successfully handled.
    stream : bool (default True)
        If `False`, the response content will be immediately downloaded.

    Returns
    -------
    bool
        `True` if the file was successfully downloaded, `False` otherwise
    """

    # Download data within 5 minutes
    with requests.get(url, stream = stream, allow_redirects = True,
            timeout = 300) as response: # type: requests.Response
        # If cannot grab file, return False
        if not response.ok:
            return False

        # Get filename
        filename: str | None = None

        ## Lookup filename from content disposition if present
        if _CONTENT_DISPOSITION in response.headers:
            if 'filename*' in (disposition := response.headers[_CONTENT_DISPOSITION]):
                filename: str = re.findall(_FILENAME_STAR_PARAM_REGEX, disposition)[0]
            elif 'filename' in disposition:
                matches: Tuple[str, str] = re.findall(_FILENAME_PARAM_REGEX, disposition)[0]
                filename: str = matches[0] if matches[0] else matches[1]

        # Set to basename of path if not present
        if not filename:
            filename = os.path.basename(urlparse(url).path)

        # Check to see if extension is present
        name, ext = os.path.splitext(filename)

        # If no extension is present and we have access to the content type
        if not ext and _CONTENT_TYPE in response.headers:
            filename = name + guess_extension(response.headers[_CONTENT_TYPE]
                .partition(';')[0].strip())

        # Handle the result of the downloaded file
        return handler(response, filename)

def download_and_write(url: str, unzip_file: bool = True, out_dir: str = os.curdir,
        stream: bool = True) -> bool:
    """Downloads a file from the specified url via a GET request and writes or unzips
    the file, if applicable.

    Parameters
    ----------
    url : str
        The url to download the file from.
    unzip_file : bool (default True)
        If `True`, will attempt to unzip the file if the file extension is correct.
    out_dir : str (default '.')
        The directory to write or unzip the file to.
    stream : bool (default True)
        If `False`, the response content will be immediately downloaded.

    Returns
    -------
    bool
        `True` if the file was successfully downloaded, `False` otherwise
    """

    def __write(__response: requests.Response, __filename: str, __dir: str) -> bool:
        """Writes the file or unzips it to the specified directory.

        Parameters
        ----------
        __response : requests.Response
            The response of the url request.
        __filename : str
            The name of the file requested from the url.
        __dir : str
            The directory to write the file(s) to.
        
        Returns
        -------
        bool
            `True` if the download was successful, `False` otherwise.
        """

        # Unzip file if available and set
        if unzip_file and __filename.endswith('.zip'):
            with BytesIO(__response.content) as zip_bytes: # type: BytesIO
                unzip(zip_bytes, out_dir = __dir)
        # Otherwise do normal extraction
        else:
            # Create directory name if not already present
            name: str = os.sep.join([__dir, __filename])
            os.makedirs(os.path.dirname(name), exist_ok = True)

            with open(name, 'wb') as file:
                for data in __response.iter_content(1024): # type: ReadableBuffer
                    file.write(data)
        return True

    return download_file(url,
        lambda response, filename: __write(response, filename, out_dir), stream = stream)



"""A script containing the methods needed for command line integration.
"""

import os
from typing import List
import click
from tomlkit import load as load_toml
import jammies.workspace.project as wspc
from jammies.defn.registrar import setup as setup_registrar
from jammies.defn.metadata import ProjectMetadata
from jammies.config import JammiesConfig, load_config, config_loc as cloc
from jammies.log import Logger

# TODO: REDO

@click.group()
def main() -> None:
    """A command line interface to construct,
    manage, and patch projects.
    """

@main.group()
def config() -> None:
    """Helpers to generate, read, and write to a config.
    """

def _check_project_config(logger: Logger, dirpath: str) -> bool:
    """Returns whether a project exists in the current location.
    
    Parameters
    ----------
    logger : Logger
        A logger for reporting on information.
    dirpath : str
        The current directory to check the path for.
    
    Returns
    -------
    bool
        `True`, if a project exists in the current location.
    """
    # If project doesn't exist, throw an error
    if not os.path.exists(
        prj_path := os.sep.join([dirpath, wspc.PROJECT_METADATA_NAME])
    ):
        logger.error(
            f'No project exists at \'{prj_path}\'. ' \
            + 'Create a project using one of the following commands:',
            '-> jammies project init',
            '-> jammies patch   init',
            sep = '\n'
        )
        return False
    
    return True

@config.command(name = 'create')
@click.option('--project', '-p', is_flag = True,
    help = 'Generates a config for the current project.')
@click.option('--site', '-s', is_flag = True,
    help = 'Generates a config for the set environment variable, ' \
    + 'virtual environment, or user if neither are specified.')
@click.option('--user', '-u', is_flag = True, help = 'Generates a config for the current user.')
@click.option('--global', '-g', '_global', is_flag = True, help = 'Generates a global config.')
@click.option('--verbose', '-v', is_flag = True, help = 'When \'true\', displays debug messages.')
def config_create(project: bool = False, site: bool = False,
        user: bool = False, _global: bool = False, verbose: bool = False) -> None:
    """Creates a configuration for the specified scope if it doesn't already
    exist. If no scope is specified, a config will be generated for the project
    scope.
    """
    # Create config object to write
    logger: Logger = Logger(verbose = verbose)
    prj_config: JammiesConfig = JammiesConfig()
    setup_registrar(logger, prj_config)

    configs_written: bool = False

    # Check if project is wanted or no config option is specified
    if project or not (project or site or user or _global):
        # If project doesn't exist, throw an error
        if not _check_project_config(logger, prj_config.dirpath):
            return

        # Otherwise check if a config does not already exist
        if os.path.exists(config_path := cloc(dirpath = prj_config.dirpath, scope = 0)):
            logger.skip(f'Config exists within project \'{config_path}\'')
        else:
            logger.debug(f'Creating config at project \'{config_path}\'')
            prj_config.write_config(scope = 0)
            configs_written = True

    # Check if site
    if site:
        # Check if a config does not already exist
        if os.path.exists(config_path := cloc(dirpath = prj_config.dirpath, scope = 1)):
            logger.skip(f'Config exists within site \'{config_path}\'')
        else:
            logger.debug(f'Creating config at site \'{config_path}\'')
            prj_config.write_config(scope = 1)
            configs_written = True

    # Check if user
    if user:
        # Check if a config does not already exist
        if os.path.exists(config_path := cloc(dirpath = prj_config.dirpath, scope = 2)):
            logger.skip(f'Config exists within user \'{config_path}\'')
        else:
            logger.debug(f'Creating config at user \'{config_path}\'')
            prj_config.write_config(scope = 2)
            configs_written = True

    # Check if global
    if _global:
        # Check if a config does not already exist
        if os.path.exists(config_path := cloc(dirpath = prj_config.dirpath, scope = 3)):
            logger.skip(f'Skip: Config exists within global \'{config_path}\'')
        else:
            logger.debug(f'Creating config at global \'{config_path}\'')
            prj_config.write_config(scope = 3)
            configs_written = True

    if configs_written:
        logger.success('Configs have been generated!')
    else:
        logger.skip('Configs are already generated!')

@config.command(name = 'loc')
@click.option('--open', '-o', 'open_dir', is_flag = True,
    help = 'If the config is found, the directory will be opened in the native file system.')
@click.option('--scope', '-s',
    type = click.Choice(['project', 'site', 'user', 'global'],
        case_sensitive = False
    ),
    default = 'project',
    help = 'The configuration to look for. If none is specified, ' \
    + 'it will default to the project config.'
)
def config_loc(open_dir: bool = False, scope: str = 'project') -> None:
    """Returns the location of the config file, if it exists."""

    logger: Logger = Logger()
    scope_val: int = int(scope == 'project') * 0 \
        + int(scope == 'site') * 1 \
        + int(scope == 'user') * 2 \
        + int(scope == 'global') * 3

    # Skip execution if an error is thrown
    if scope_val == 0 and not _check_project_config(logger, os.curdir):
        return

    # Check if the current config path exists
    if os.path.exists(config_path := cloc(scope = scope_val)):
        logger.success(config_path)

        # Open file location if available
        if open_dir:
            os.startfile(os.path.dirname(config_path))
    else:
        logger.error(
            f'No config for {scope}. Create the config using:',
            f'-> jammies config create --{scope}',
            sep = '\n'
        )

@config.command(name = 'list')
def config_list() -> None:
    """Lists all available configuration options."""
    logger: Logger = Logger()
    prj_config: JammiesConfig = JammiesConfig()
    top_message: List[str] = ['Available config options']
    top_message += map(lambda s: f'-> {s}', prj_config.list_vals())
    logger.success(
        *top_message,
        sep = '\n'
    )

@config.command(name = 'value')
@click.argument('name')
@click.argument('value', required = False)
@click.option('--scope', '-s',
    type = click.Choice(['project', 'site', 'user', 'global'],
        case_sensitive = False
    ),
    default = 'project',
    help = 'The configuration to look for. If none is specified, ' \
    + 'it will default to the project config.'
)
def config_value(name: str, value: str | None = None, scope: str = 'project') -> None:
    """Gets the configuration value associated with the name in the
    specified scope. If the value is specified, the name will be updated
    to hold that value.
    """
    logger: Logger = Logger()
    scope_val: int = int(scope == 'project') * 0 \
        + int(scope == 'site') * 1 \
        + int(scope == 'user') * 2 \
        + int(scope == 'global') * 3

    # Skip execution if project doesn't exist
    if scope_val == 0 and not _check_project_config(logger, os.curdir):
        return

    # Otherwise check if config exists
    if not os.path.exists(config_path := cloc(scope = scope_val)):
        logger.error(
            f'No config for {scope}. Create the config using:',
            f'-> jammies config create --{scope}',
            sep = '\n'
        )
        return

    # Load config scope
    prj_config: JammiesConfig = None
    with open(config_path, mode = 'r', encoding = 'UTF-8') as file:
        prj_config = load_toml(file)
        prj_config['dirpath'] = os.curdir
        prj_config = JammiesConfig.decode_toml(prj_config)

    # If a value is present, set it within the config
    if value:
        success, val = prj_config.set_val(name, value)
        if success:
            prj_config.write_config(scope = scope_val)
            logger.success(f'[{scope}] {name}: {val}')
        else:
            logger.error(f'[{scope}] {val}')
    else:
        # Otherwise, read the value
        success, val = prj_config.get_val(name)
        if success:
            logger.success(f'[{scope}] {name} -> {val}')
        else:
            logger.error(f'[{scope}] {val}')

@main.group()
def patch() -> None:
    """Helpers to patch an existing project.
    """

@patch.command(name = 'init')
@click.option(
    '--import_metadata', '-I',
    type = str,
    default = None,
    help = 'A path or URL to the metadata JSON.'
)
@click.option(
    '-a', '-A', 'include_hidden',
    is_flag = True,
    help = 'When added, copies hidden files to the working directory.'
)
def init(import_metadata: str | None = None, include_hidden: bool = False) -> None:
    """Initializes a new project or an existing project from the
    metadata JSON in the executing directory, an import, or from
    the metadata builder if neither are present.
    """

    # Read config
    logger: Logger = Logger()
    prj_config: JammiesConfig = load_config()
    setup_registrar(logger, prj_config)

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(dirpath = prj_config.dirpath,
        import_loc = import_metadata)
    clean_dir = metadata.location['clean']
    working_dir = metadata.location['src']
    patch_dir = metadata.location['patches']
    out_dir = metadata.location['out']

    # Setup workspace
    if wspc.setup_clean(metadata, logger, config = prj_config, clean_dir = clean_dir):
        wspc.setup_working(clean_dir = clean_dir, working_dir = working_dir,
            patch_dir = patch_dir, out_dir = out_dir, include_hidden = include_hidden)
        print('Initialized patching environment!')
    else:
        print('Could not generate clean workspace.')

@patch.command(name = 'output')
def output() -> None:
    """Generates any patches and clones the new files to an output
    directory."""

    # Read config
    logger: Logger = Logger()
    prj_config: JammiesConfig = load_config()
    setup_registrar(logger, prj_config)

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(dirpath = prj_config.dirpath)
    clean_dir = metadata.location['clean']
    working_dir = metadata.location['src']
    patch_dir = metadata.location['patches']
    out_dir = metadata.location['out']

    # Output working and generate patches
    wspc.output_working(metadata, clean_dir = clean_dir, working_dir = working_dir,
        patch_dir = patch_dir, out_dir = out_dir)

    print('Generated patches and output files!')

@patch.command(name = 'clean')
@click.option(
    '--import_metadata', '-I',
    type = str,
    default = None,
    help = 'A path or URL to the metadata JSON.'
)
def clean(import_metadata: str | None = None) -> None:
    """Initializes a clean workspace from the
    metadata JSON in the executing directory, an import, or from
    the metadata builder if neither are present.
    """

    # Read config
    logger: Logger = Logger()
    prj_config: JammiesConfig = load_config()
    setup_registrar(logger, prj_config)

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(dirpath = prj_config.dirpath,
        import_loc = import_metadata)
    clean_dir = metadata.location['clean']

    # Setup workspace
    if wspc.setup_clean(metadata, logger, config = prj_config, clean_dir = clean_dir):
        print('Setup clean workspace!')
    else:
        print('Could not generate clean workspace.')

@patch.command(name = 'src')
@click.option(
    '--import_metadata', '-I',
    type = str,
    default = None,
    help = 'A path or URL to the metadata JSON.'
)
def source(import_metadata: str | None = None) -> None:
    """Initializes a patched workspace from the
    metadata JSON in the executing directory, an import, or from
    the metadata builder if neither are present.
    """

    # Read config
    logger: Logger = Logger()
    prj_config: JammiesConfig = load_config()
    setup_registrar(logger, prj_config)

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(dirpath = prj_config.dirpath,
        import_loc = import_metadata)
    working_dir = metadata.location['src']
    patch_dir = metadata.location['patches']
    out_dir = metadata.location['out']

    # Setup workspace
    if wspc.setup_clean(metadata, logger, config = prj_config, clean_dir = working_dir,
            invalidate_cache = True):
        wspc.setup_working_raw(working_dir = working_dir,
            patch_dir = patch_dir, out_dir = out_dir)
        print('Setup patched workspace!')
    else:
        print('Could not generate clean workspace.')



"""A script obtaining a project file from a git repository.
"""

import os
import subprocess as sp
from jammies.module import load_module, has_module
from jammies.registrar import JammiesRegistrar

_REGISTRY_NAME: str = 'git'
"""The registry name of the project file handler."""

def setup(registrar: JammiesRegistrar) -> None:
    """A setup method used to register components to the project.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    """

    # Check if git is present on the machine
    if sp.run(['git', '--help'], check = False, stdout = sp.DEVNULL).returncode == 0 \
            and has_module('git'):
        load_module('jammies.internal.file.delegate.gitrepo').setup_delegate(registrar)
    else:
        registrar.add_file_handler_missing_message(_REGISTRY_NAME,
            'git, and/or the corresponding Python library, is not installed'
            + ' on the local machine. You can install git using the installer'
            + ' at https://git-scm.com/downloads.'
            + os.linesep + os.linesep
            + 'You can install the corresponding library by running one of the'
            + ' following commands:'
            + os.linesep
            + '-> pip install GitPython>=3.1' + os.linesep
            + '-> pip install jammies[git]'
        )



"""A script containing information about handling the project metadata.
"""

import os
from typing import List, Tuple, Set, Dict
from pathlib import Path
import shutil
import fnmatch
from jammies.log import Logger
from jammies.registrar import JammiesRegistrar
from jammies.defn.file import ProjectFile
from jammies.struct.codec import DictCodec, DictObject
from jammies.utils import get_or_default, input_yn_default
from jammies.config import JammiesConfig

_DEFAULT_LOCATIONS: Dict[str, str] = {
    'clean': 'clean',
    'src': 'src',
    'patches': 'patches',
    'out': 'out'
}
"""The default locations used by the project metadata."""

class ProjectMetadata:
    """Metadata information associated with the project being patched or ran."""

    def __init__(self, files: List[ProjectFile],
            ignore: List[str] | None = None, overwrite: List[str] | None = None,
            location: DictObject | None = None, extra: DictObject | None = None) -> None:
        """
        Parameters
        ----------
        files : list of `ProjectFile`s
            The files associated with the project.
        ignore : list of str (default '[]')
            The patterns for files that are ignored for patching.
        overwrite : list of str (default '[]')
            The patterns for files that are overwritten instead of patching.
        location : dict[str, str] | None (default 'None')
            A alias map for directories used by this project.
        extra : dict[str, Any] (default '{}')
            Extra data defined by the user.
        """
        self.files: List[ProjectFile] = files
        self.ignore: List[str] = [] if ignore is None else ignore
        self.overwrite: List[str] = [] if overwrite is None else overwrite
        self.location: Dict[str, str] = {} if location is None else location
        # Add Missing Defaults
        for key, value in _DEFAULT_LOCATIONS.items():
            if key not in self.location:
                self.location[key] = value
        self.extra: DictObject = {} if extra is None else extra

    def __copy_and_log(self, src: str, dst: str, config: JammiesConfig) -> object:
        """Copies and logs a generated file.

        Parameters
        ----------
        src : str
            The source location of the file.
        dst : str
            The destination of the file.
        config : `JammiesConfig`
            The configuration settings.
        """

        output: object = shutil.copy2(src, dst)
        config.internal.add_generated_file(dst)
        return output

    def setup(self, root_dir: str, logger: Logger, config: JammiesConfig | None = None) -> bool:
        """Sets up the project for usage.

        Parameters
        ----------
        root_dir : str
            The root directory to set up the project in.
        logger : `Logger`
            A logger for reporting on information.
        config : JammiesConfig | None (default 'None')
            The configuration settings.
        """

        # Display warning message if no config is present or the warning message is enabled
        if (not config) or config.project.display_warning_message:
            if not input_yn_default('You are about to download project files from third parties. '
                + 'jammies and its maintainers are not liable for anything that happens '
                + 'as a result of downloading or using these files. Would you still like to '
                + 'download these files?', True):
                return False
            # Ask to disable warning message if config is present
            if config:
                def _disable_warning_message(conf: JammiesConfig) -> None:
                    """Disables the warning message within the config.
                    
                    Parameters
                    ----------
                    config : JammiesConfig
                        The configuration settings.
                    """
                    conf.project.display_warning_message = False

                config.update_and_write(_disable_warning_message,
                    save = input_yn_default(
                        'Would you like to hide this warning message from now on?',
                        False
                    )
                )

        # Clear generated files
        if config:
            config.internal.clear_generated_files()

        failed: List[ProjectFile] = []
        tmp_root: str = os.path.join(root_dir, '.tmp')

        for file in self.files: # type: ProjectFile
            # Setup file
            if not file.setup(tmp_root, ignore_sub_directory = True):
                logger.error(f'Failed to setup {file.registry_name()}')
                failed.append(file)
                shutil.rmtree(tmp_root)
                continue

            # Apply post processor
            if file.post_processor and not file.post_processor[0](logger, tmp_root):
                pp_name: str = file.post_processor[1]['type']
                logger.error(f'Failed to apply post processor {pp_name} to {file.registry_name()}')
                failed.append(file)
                shutil.rmtree(tmp_root)
                continue

            shutil.copytree(tmp_root, file.create_path(root_dir),
                copy_function=lambda src, dst: self.__copy_and_log(src, dst, config) \
                    if config else shutil.copy2,
                dirs_exist_ok=True
            )
            shutil.rmtree(tmp_root)

        # Write generated files
        if config:
            config.update_and_write(lambda config: None, save = True)

        # Verify no files failed at any point
        return not failed

    def codec(self) -> 'ProjectMetadataCodec':
        """Returns the codec used to encode and decode this metadata.

        Returns
        -------
        ProjectFileCodec
            The codec used to encode and decode this metadata.
        """
        return METADATA_CODEC

    def ignore_and_overwrite(self, root_dir: str) -> Tuple[Set[str], Set[str]]:
        """Gets the ignored and overwritten files from the specified directory.

        Parameters
        ----------
        root_dir : str
            The root directory to get the files of.
        
        Returns
        -------
        (set of strs, set of strs)
            A tuple of ignored and overwritten files, respectively.
        """

        # Get all files
        files: list[str] = [Path(os.path.relpath(path, root_dir)).as_posix() \
            for path in Path(root_dir).rglob('*') if path.is_file()]

        # Get ignored files
        ignored_files: Set[str] = set()
        for ignore in self.ignore: # type: str
            ignored_files.update(fnmatch.filter(files, ignore))

        # Get overwritten files
        overwritten_files: Set[str] = set()
        for overwrite in self.overwrite: # type: str
            overwritten_files.update(fnmatch.filter(files, overwrite))

        return (ignored_files, overwritten_files)

def build_metadata() -> ProjectMetadata:
    """Builds a ProjectMetadata from user input.
    
    Returns
    -------
    ProjectMetadata
        The built project metadata.
    """

    # Project Files
    available_file_types: str = ', '.join(METADATA_CODEC.registrar.get_available_builders())
    files: List[ProjectFile] = []
    flag: bool = True
    while flag:
        file_type: str = input(f'Add file ({available_file_types}): ').lower()
        files.append(
            METADATA_CODEC.registrar.get_project_file_builder(file_type)(METADATA_CODEC.registrar)
        )
        flag = input_yn_default('Would you like to add another file', True)

    # Ignore Patterns
    flag = input_yn_default('Would you like to ignore any files when patching', False)
    ignore: List[str] = []
    while flag:
        ignore.append(input('Add pattern to ignore: '))
        flag = input_yn_default('Would you like to ignore another pattern', True)

    # Overwrite Patterns
    flag = input_yn_default('Would you like to overwrite any files when patching', False)
    overwrite: List[str] = []
    while flag:
        overwrite.append(input('Add pattern to overwrite: '))
        flag = input_yn_default('Would you like to overwrite another pattern', True)

    # Create metadata
    return ProjectMetadata(files, ignore = ignore, overwrite = overwrite)

class ProjectMetadataCodec(DictCodec[ProjectMetadata]):
    """A codec for encoding and decoding a ProjectMetadata.
    """

    def __init__(self) -> None:
        """
        Parameters
        ----------
        registrar : `JammiesRegistrar`
            The registrar used to register the components for the project.
        """
        self.registrar: JammiesRegistrar = None

    def encode(self, obj: ProjectMetadata) -> DictObject:
        dict_obj: DictObject = {}
        dict_obj['files'] = list(map(lambda file: file.codec().encode(file), obj.files))
        if obj.ignore:
            dict_obj['ignore'] = obj.ignore
        if obj.overwrite:
            dict_obj['overwrite'] = obj.overwrite

        location: DictObject = {}
        for key, value in obj.location.items():
            # If the key does not have a default or is not the default value
            if key not in _DEFAULT_LOCATIONS or _DEFAULT_LOCATIONS[key] != value:
                location[key] = value
        if location:
            dict_obj['location'] = location

        if obj.extra:
            dict_obj['extra'] = obj.extra
        return dict_obj

    def __decode_file(self, file: DictObject) -> ProjectFile:
        """Decodes a project file from its type.

        Parameters
        ----------
        file : Dict[str, Any]
            The encoded project file.
        
        Returns
        -------
        `ProjectFile`
            The decoded project file.
        """
        return self.registrar.get_project_file_codec(file['type']).decode(file)

    def decode(self, obj: DictObject) -> ProjectMetadata:
        return ProjectMetadata(list(map(self.__decode_file, obj['files'])),
            ignore = get_or_default(obj, 'ignore', ProjectMetadata),
            overwrite = get_or_default(obj, 'overwrite', ProjectMetadata),
            location = get_or_default(obj, 'location', ProjectMetadata),
            extra = get_or_default(obj, 'extra', ProjectMetadata))

METADATA_CODEC: ProjectMetadataCodec = ProjectMetadataCodec()
"""The codec for :class:`jammies.metadata.base.ProjectMetadata`."""



"""A script containing the post processor which unpacks archives
in the directory they currently exist within.
"""

import os
from glob import glob
from shutil import unpack_archive, _find_unpack_format
from jammies.log import Logger
from jammies.registrar import JammiesRegistrar

_REGISTRY_NAME: str = 'unpack_archive'
"""The registry name of the post processor."""

def setup(registrar: JammiesRegistrar) -> None:
    """A setup method used to register components to the project.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    """
    registrar.register_post_processor(_REGISTRY_NAME, unpack)

def unpack(logger: Logger, current_dir: str, **kwargs) -> bool:
    """Unpacks an archive when present and removes
    the original archive.
    
    Parameters
    ----------
    logger : Logger
        A logger for reporting on information.
    current_dir : str
        The current directory to execute the processor within.
    recursive : bool (default `False`)
        When `True`, recurses through subdirectories in the current
        directory.
    
    Returns
    -------
    bool
        `True` if the operation is successful.
    """
    archives: list[str] = []

    logger.debug(f'Unpacking archives in {current_dir}')

    for filename in glob(os.sep.join([current_dir, '**']),
            recursive = kwargs['recursive'] if 'recursive' in kwargs else False):
        # Check if file is an archive
        if os.path.isfile(filename) and _find_unpack_format(filename):
            archives.append(filename)

    for filename in archives:
        logger.debug(f'Found: {filename}')
        # Unpack and delete archive
        unpack_archive(filename, extract_dir = os.path.dirname(filename))
        os.remove(filename)

    # Return true if an unpacking was performed
    return len(archives) > 0



"""A script obtaining a project file from a download url.
"""

from jammies.struct.codec import DictObject
from jammies.defn.file import ProjectFile, ProjectFileCodec
from jammies.defn.builder import build_file
from jammies.utils import download_and_write
from jammies.registrar import JammiesRegistrar

def setup(registrar: JammiesRegistrar) -> None:
    """A setup method used to register components to the project.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    """

    # Create codec
    codec: URLProjectFileCodec = URLProjectFileCodec(registrar)

    registrar.register_file_handler(_REGISTRY_NAME, codec, build_url)

_REGISTRY_NAME: str = 'url'
"""The registry name of the project file handler."""

class URLProjectFile(ProjectFile):
    """A project file for a file at a downloadable url link.
    The file will be obtained via a GET request."""

    def __init__(self, url: str, **kwargs) -> None:
        """
        Parameters
        ----------
        url : str
            The downloadable link for the file.
        """
        super().__init__(**kwargs)
        self.url: str = url

    def registry_name(self) -> str:
        return _REGISTRY_NAME

    def setup(self, root_dir: str, ignore_sub_directory: bool = False) -> bool:
        super().setup(root_dir, ignore_sub_directory = ignore_sub_directory)
        base_path: str = root_dir if ignore_sub_directory else self.create_path(root_dir)
        return download_and_write(self.url, unzip_file = False,
            out_dir = base_path)

def build_url(registrar: JammiesRegistrar) -> URLProjectFile:
    """Builds an URLProjectFile from user input.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.

    Returns
    -------
    URLProjectFile
        The built project file.
    """
    url: str = input('Direct URL: ')
    return build_file(registrar, _REGISTRY_NAME, lambda kwargs: URLProjectFile(url, **kwargs))

class URLProjectFileCodec(ProjectFileCodec[URLProjectFile]):
    """A codec for encoding and decoding an URLProjectFile.
    """

    def encode_type(self, obj: URLProjectFile, dict_obj: DictObject) -> DictObject:
        dict_obj['url'] = obj.url
        return dict_obj

    def decode_type(self, obj: DictObject, **kwargs: DictObject) -> URLProjectFile:
        kwargs['codec'] = self # Set codec
        return URLProjectFile(obj['url'], **kwargs)



"""A script containing the definition of a registrar used to register
the necessary components of the project.
"""
from typing import TypeAlias, Callable
from abc import ABC, abstractmethod
from jammies.defn.file import ProjectFileCodec, ProjectFile, PostProcessor

ProjectFileBuilder: TypeAlias = Callable[['JammiesRegistrar'], ProjectFile]
"""A supplier used to construct a ProjectFile from an user's input."""

class JammiesRegistrar(ABC):
    """A registrar for the project. Provides endpoints to register file
    handlers and post processors.
    """

    def __init__(self) -> None:
        """"""

    @abstractmethod
    def register_file_handler(self, name: str, codec: ProjectFileCodec,
            builder: ProjectFileBuilder) -> None:
        """Registers a `ProjectFile` handler. A handler is used to pull file
        from a given location.
        
        Parameters
        ----------
        name : str
            The name of the file handler to register.
        codec : `ProjectFileCodec`
            The codec used to encode and decode the `ProjectFile`.
        builder : `ProjectFileBuilder`
            The builder used to construct the project file within a
            shell.
        """

    @abstractmethod
    def add_file_handler_missing_message(self, name: str, message: str) -> None:
        """Adds a message for a file handler if it is referenced without
        being registered. A message should be registered as an alternative
        to a file handler.
        
        Parameters
        ----------
        name : str
            The name of the file handler to register.
        message : str
            The message to display when not present.
        """

    @abstractmethod
    def register_post_processor(self, name: str, post_processor: PostProcessor) -> None:
        """Registers a post processor. A post processor is used to apply
        additional changes to a `ProjectFile` after it has been pulled.
        
        Parameters
        ----------
        name : str
            The name of the post processor to register.
        post_processor : `PostProcessor`
            The post processor to apply to a project file.
        """

    @abstractmethod
    def add_post_processor_missing_message(self, name: str, message: str) -> None:
        """Adds a message for a post processor if it is referenced without
        being registered. A message should be registered as an alternative
        to a post processor.
        
        Parameters
        ----------
        name : str
            The name of the post processor to register.
        message : str
            The message to display when not present.
        """

    @abstractmethod
    def get_project_file_codec(self, name: str) -> ProjectFileCodec:
        """Gets a `ProjectFileCodec` for the associated name.

        Parameters
        ----------
        name : str
            The name of the codec.

        Returns
        -------
        `ProjectFileCodec`
            The registered codec.
        """

    @abstractmethod
    def get_project_file_builder(self, name: str) -> ProjectFileBuilder:
        """Gets a `ProjectFileBuilder` for the associated name.

        Parameters
        ----------
        name : str
            The name of the builder.

        Returns
        -------
        `ProjectFileBuilder`
            The registered builder.
        """

    @abstractmethod
    def get_post_processor(self, name: str) -> PostProcessor:
        """Gets a `PostProcessor` for the associated name.

        Parameters
        ----------
        name : str
            The name of the post processor.

        Returns
        -------
        `PostProcessor`
            The registered post processor.
        """

    @abstractmethod
    def get_available_builders(self) -> list[str]:
        """Returns a list of the available project file builders.
        
        Returns
        -------
        list of str
            A list of project file builder names.
        """



"""A script which holds all lazy references to a given
class."""

import sys
from types import ModuleType
from typing import Callable
from importlib.machinery import ModuleSpec
from importlib.util import find_spec, LazyLoader, module_from_spec, spec_from_file_location

def has_module(name: str) -> bool:
    """Checks whether the module is currently loaded or can be added to the current workspace.

    Parameters
    ----------
    name : str
        The name of the module.

    Returns
    -------
    bool
        `True` if the module exists, `False` otherwise.
    """
    return (name in sys.modules) or (find_spec(name) is not None)

def load_module(module_name: str,
        spec_getter: Callable[[str], ModuleSpec] = find_spec) -> ModuleType:
    """Loads a module based on its name and getter for the spec.
    If the module is already loaded, it will be used instead.

    Parameters
    ----------
    module_name : str
        The name of the module to load.
    spec_getter : (str) -> ModuleSpec (default 'importlib.util.find_spec')
        Determines how to load the spec with the specified module name.
    
    Returns
    -------
    ModuleType
        The loaded module.
    """
    # If already loaded, just return the module
    if module_name in sys.modules:
        return sys.modules[module_name]

    # Load module if not present
    spec: ModuleSpec = spec_getter(module_name)
    module: ModuleType = module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module

def lazy_import(name: str) -> ModuleType:
    """Sets up the module to be lazily imported if it is not already loaded.

    Parameters
    ----------
    name : str
        The name of the module to lazily load.
    
    Returns
    -------
    types.ModuleType
        The module which will be loaded on first execution.
    """

    def _set_lazy(module_name: str) -> ModuleSpec:
        """Sets the spec loader to load the module lazily.

        Parameters
        ----------
        module_name : str
            The name of the module to lazily load.
        
        Returns
        -------
        ModuleSpec
            The lazily loaded spec.
        """

        # Make spec lazily loaded
        spec: ModuleSpec | None = find_spec(module_name)
        loader: LazyLoader = LazyLoader(spec.loader)
        spec.loader = loader
        return spec

    return load_module(name, spec_getter = _set_lazy)

def dynamic_import(module_type: str, name: str, path: str) -> ModuleType:
    """Dynamically imports a module into the current Python executable.
    Modules dynamically imported will be prefixed with `jammies.dynamic`,
    followed by the module type and the associated name.

    Parameters
    ----------
    module_type : str
        The type of module being dynamically imported.
    name : str
        The name of the module to import.
    path : str
        The location of the module to import.
    
    Returns
    -------
    ModuleType
        The dynamically loaded module.
    """
    module_name: str = f'jammies.dynamic.{module_type}.{name}'

    return load_module(module_name,
        spec_getter = lambda mdn: spec_from_file_location(mdn, location = path))



"""A script containing the implemented registrar used within the project."""

from typing import Dict, Any

from jammies.log import Logger
from jammies.config import JammiesConfig
from jammies.struct.registry import Registry
from jammies.registrar import JammiesRegistrar, ProjectFileBuilder
from jammies.defn.file import ProjectFileCodec, PostProcessor
from jammies.defn.metadata import METADATA_CODEC

import jammies.internal.file.osf as file_osf
import jammies.internal.file.url as file_url
import jammies.internal.file.gitrepo as file_git
import jammies.internal.postprocessor.unpack as post_unpack
import jammies.internal.postprocessor.notebook as post_notebook

class RegistryError(Exception):
    """Raised when an error occurs when operating on a registry."""

class JammiesRegistrarImpl(JammiesRegistrar):
    """An implementation of the registrar class used to hold
    the registered instances of the project.
    """

    def __init__(self) -> None:
        """"""

        # Initialize Stager
        # 0: Internal Logic
        # 1: Dynamic Script Logic
        # 2: Frozen
        self.__stager: int = 0

        # Project File Handlers
        self.file_codecs: Registry[ProjectFileCodec] = Registry()
        self.file_builders: Registry[ProjectFileBuilder] = Registry()
        self.file_handler_exceptions: Registry[str] = Registry()

        # Post Processors
        self.post_processors: Registry[PostProcessor] = Registry()
        self.post_processor_exceptions: Registry[str] = Registry()

    def __check_preconditions(self, name: str, registry: Dict[str, Any],
            registry_type: str) -> None:
        """Checks the preconditions necessary to access a registry method.
        
        Parameters
        ----------
        name : str
            The name of the object to register.
        registry : dict
            The registry to registry the object to.
        registry_type : str
            A string representation of the registry data.
        """

        if self.__stager > 1:
            raise RegistryError(f'{name} could not be registered; the registry has been frozen.')
        if self.__stager > 0 and ':' not in name:
            raise ValueError(f'{name} must contain a \':\', '\
                + 'where the prefix represents a unique identifier for the group of scripts.')
        if name in registry:
            raise ValueError(f'{name} already has a registered {registry_type}.')

    def register_file_handler(self, name: str, codec: ProjectFileCodec,
            builder: ProjectFileBuilder) -> None:
        self.__check_preconditions(name, self.file_codecs, 'file handler')

        self.file_codecs[name] = codec
        self.file_builders[name] = builder

    def add_file_handler_missing_message(self, name: str, message: str) -> None:
        self.__check_preconditions(name, self.file_handler_exceptions, 'message')
        if name in self.file_codecs:
            raise ValueError(f'{name} is registered, a message handler is not needed.')

        self.file_handler_exceptions[name] = message

    def register_post_processor(self, name: str, post_processor: PostProcessor) -> None:
        self.__check_preconditions(name, self.post_processors, 'post processor')

        self.post_processors[name] = post_processor

    def add_post_processor_missing_message(self, name: str, message: str) -> None:
        self.__check_preconditions(name, self.post_processor_exceptions, 'message')
        if name in self.post_processors:
            raise ValueError(f'{name} is registered, a message handler is not needed.')

        self.post_processor_exceptions[name] = message

    def get_project_file_codec(self, name: str) -> ProjectFileCodec:
        return self.file_codecs[name]

    def get_project_file_builder(self, name: str) -> ProjectFileBuilder:
        return self.file_builders[name]

    def get_post_processor(self, name: str) -> PostProcessor:
        return self.post_processors[name]

    def get_available_builders(self) -> list[str]:
        return self.file_builders.keys()

    def stage(self) -> None:
        """Stages the registrar into its next state.
        """
        self.__stager += 1

REGISTRAR: JammiesRegistrar = JammiesRegistrarImpl()

def setup(logger: Logger, config: JammiesConfig) -> None:
    """Initializes the information within the registrar.

    Parameters
    ----------
    logger : `Logger`
        A logger for reporting on information.
    config : `JammiesConfig`
        The configuration settings.
    """

    # Setup basic values
    METADATA_CODEC.registrar = REGISTRAR

    # Register internal implementations
    logger.debug('Registering internal implementations...')
    file_osf.setup(REGISTRAR)
    file_url.setup(REGISTRAR)
    file_git.setup(REGISTRAR)
    post_unpack.setup(REGISTRAR)
    post_notebook.setup(REGISTRAR)

    REGISTRAR.stage()

    # Register dynamic implementations
    logger.debug('Registering dynamic implementations...')
    loaded_scripts: int = config.load_dynamic_scripts(logger, REGISTRAR)
    if loaded_scripts > 0:
        logger.debug(f'Successfully loaded {loaded_scripts} scripts')

    REGISTRAR.stage()



"""A script obtaining a project file from an Open Science Framework
project.
"""

from jammies.struct.codec import DictObject
from jammies.defn.file import ProjectFile, ProjectFileCodec
from jammies.defn.builder import build_file
from jammies.utils import download_and_write
from jammies.registrar import JammiesRegistrar

def setup(registrar: JammiesRegistrar) -> None:
    """A setup method used to register components to the project.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    """

    # Create codec
    codec: OSFProjectFileCodec = OSFProjectFileCodec(registrar)

    registrar.register_file_handler(_REGISTRY_NAME, codec, build_osf)

_REGISTRY_NAME: str = 'osf'
"""The registry name of the project file handler."""

class OSFProjectFile(ProjectFile):
    """A project file for an Open Science Framework repository."""

    def __init__(self, project_id: str, **kwargs: DictObject) -> None:
        """
        Parameters
        ----------
        project_id : str
            The five character identifier of the repository.
        """
        super().__init__(**kwargs)
        self.project_id: str = project_id
        self.__url: str = \
            f'https://files.osf.io/v1/resources/{project_id}/providers/osfstorage/?zip='

    def registry_name(self) -> str:
        return _REGISTRY_NAME

    def setup(self, root_dir: str, ignore_sub_directory: bool = False) -> bool:
        super().setup(root_dir, ignore_sub_directory = ignore_sub_directory)
        base_path: str = root_dir if ignore_sub_directory else self.create_path(root_dir)
        return download_and_write(self.__url, out_dir = base_path)

def build_osf(registrar: JammiesRegistrar) -> OSFProjectFile:
    """Builds an OSFProjectFile from user input.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.

    Returns
    -------
    OSFProjectFile
        The built project file.
    """
    project_id: str = input('OSF Project Id: ')
    return build_file(registrar, _REGISTRY_NAME,
        lambda kwargs: OSFProjectFile(project_id, **kwargs)
    )

class OSFProjectFileCodec(ProjectFileCodec[OSFProjectFile]):
    """A codec for encoding and decoding an OSFProjectFile.
    """

    def encode_type(self, obj: OSFProjectFile, dict_obj: DictObject) -> DictObject:
        dict_obj['id'] = obj.project_id
        return dict_obj

    def decode_type(self, obj: DictObject, **kwargs: DictObject) -> OSFProjectFile:
        kwargs['codec'] = self # Set codec
        return OSFProjectFile(obj['id'], **kwargs)



"""A script obtaining a project file from a git repository.
"""

import os
from typing import Set
from git import Repo
from git.util import rmtree
from jammies.utils import get_default, input_with_default, input_yn_default
from jammies.struct.codec import DictObject
from jammies.defn.file import ProjectFile, ProjectFileCodec
from jammies.defn.builder import build_file
from jammies.registrar import JammiesRegistrar
from jammies.internal.file.gitrepo import _REGISTRY_NAME

def setup_delegate(registrar: JammiesRegistrar) -> None:
    """A setup method used to register components to the project.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    """

     # Create codec
    codec: GitProjectFileCodec = GitProjectFileCodec(registrar)

    registrar.register_file_handler(_REGISTRY_NAME, codec, build_git)

_VALID_BRANCH_TYPES: Set[str] = {
    'branch',
    'commit',
    'tag'
}
"""A set of valid keys indicating the checkout location of the Git repository."""

class GitProjectFile(ProjectFile):
    """A project file for an Git repository."""

    def __init__(self, repository: str, branch_type: str = 'branch',
            branch: str | None = None, **kwargs: DictObject) -> None:
        """
        Parameters
        ----------
        repository : str
            The Git link for the repository location.
        branch_type : str (default 'branch')
            The name of the key holding the branch. Must be within `_VALID_BRANCH_TYPES`.
        branch : str | None (default None)
            The name of the checkout location. If `None`, the default checkout
            location will be used.
        """
        super().__init__(**kwargs)
        self.repository: str = repository
        self.branch: str | None = branch
        if branch_type not in _VALID_BRANCH_TYPES:
            raise ValueError(f"'{branch_type}' is not a valid branch type. "
                + f"Specify one of the following: {', '.join(_VALID_BRANCH_TYPES)}")
        self.branch_type: str = branch_type

    def registry_name(self) -> str:
        return _REGISTRY_NAME

    def setup(self, root_dir: str, ignore_sub_directory: bool = False) -> bool:
        super().setup(root_dir, ignore_sub_directory = ignore_sub_directory)
        base_path: str = root_dir if ignore_sub_directory else self.create_path(root_dir)

        # Checkout and change branches, if applicable
        with Repo.clone_from(self.repository, base_path) as repo:
            if self.branch is not None:
                repo.git.checkout(self.branch)

        rmtree(os.path.join(base_path, '.git'))
        return True

def build_git(registrar: JammiesRegistrar) -> GitProjectFile:
    """Builds a GitProjectFile from user input.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.

    Returns
    -------
    GitProjectFile
        The built project file.
    """
    repository: str = input('Git Repository: ')
    if input_yn_default('Would you like to specify a checkout location', True):
        branch_type: str = input_with_default(
            GitProjectFile, 'branch_type', ', '.join(_VALID_BRANCH_TYPES))
        branch: str | None = input(f'{branch_type.capitalize()} id: ')
    else:
        branch_type: str = get_default(GitProjectFile, 'branch_type')
        branch: str | None = None
    return build_file(registrar, _REGISTRY_NAME, lambda kwargs:
        GitProjectFile(repository, branch = branch, branch_type = branch_type, **kwargs)
    )

class GitProjectFileCodec(ProjectFileCodec[GitProjectFile]):
    """A codec for encoding and decoding a GitProjectFile.
    """

    def encode_type(self, obj: GitProjectFile, dict_obj: DictObject) -> DictObject:
        dict_obj['repository'] = obj.repository
        if obj.branch != get_default(GitProjectFile, 'branch'):
            dict_obj[obj.branch_type] = obj.branch
        return dict_obj

    def decode_type(self, obj: DictObject, **kwargs: DictObject) -> GitProjectFile:
        kwargs['codec'] = self # Set codec
        for branch_name in _VALID_BRANCH_TYPES: # type: str
            if branch_name in obj:
                return GitProjectFile(
                    obj['repository'],
                    branch = obj[branch_name],
                    branch_type = branch_name,
                    **kwargs
                )
        return GitProjectFile(obj['repository'], **kwargs)



"""A script holding logger methods for the module.
"""

from typing import List

def _log(header: str, *args, **kwargs) -> None:
    """A general logger which applies a header
    to the first message. All other parameters
    are the same as `print`.
    
    Parameters
    ----------
    header : str
        The header to prefix to the first message.
    """
    args: List = list(args)
    args[0] = f'{header}: {args[0]}'
    print(*args, **kwargs)

class Logger:
    """A general logger for messages in the module."""

    def __init__(self, verbose: bool = True) -> None:
        """
        Parameters
        ----------
        verbose : bool
            Whether debug messages should be logged.
        """
        self.verbose: bool = verbose

    def skip(self, *args, **kwargs) -> None:
        """Prefixes 'Skip' in front of a message."""
        _log('Skip', *args, **kwargs)

    def error(self, *args, **kwargs) -> bool:
        """Prefixes 'Error' in front of a message."""
        _log('Error', *args, **kwargs)
        return False

    def success(self, *args, **kwargs) -> bool:
        """Prefixes 'Success' in front of a message."""
        _log('Success', *args, **kwargs)
        return True

    def debug(self, *args, **kwargs) -> None:
        """Prefixes 'Debug' in front of a message if
        verbose is enabled."""
        if self.verbose:
            _log('Debug', *args, **kwargs)



"""A script containing the configuration information.
"""

import sys
import os
from glob import iglob
from types import ModuleType
from typing import Callable, Any, Tuple, List, Set
from platformdirs import site_config_dir, user_config_dir
from tomlkit import table, document, comment, TOMLDocument, load, dump, boolean, array
from tomlkit.items import Table, Array
from jammies.log import Logger
from jammies.registrar import JammiesRegistrar
from jammies.struct.codec import DictObject
from jammies.module import dynamic_import, load_module

_ENV_VAR: str = 'JAMMIES_CONFIG_DIR'
"""The environment variable pointing to the config directory."""

_CONFIG_DIR: str = 'jammies'
"""The config directory for jammies."""

_SCRIPT_DIR: str = 'scripts'
"""The script directory for jammies."""

_CONFIG_FILE: str = 'jammies.toml'
"""The name of the config file."""

def _project_config(dirpath: str, path : str) -> str:
    """Returns the path relative to the project configuration.
    
    Parameters
    ----------
    dirpath : str
        The root directory of the project.
    path : str
        The relativized path.
        
    Returns
    -------
    str
        The path relative to the project configuration.
    """
    return os.sep.join([dirpath, path])

def _env_var_config(path : str) -> str | None:
    """Returns the path relative to the environment variable configuration.
    
    Parameters
    ----------
    path : str
        The relativized path.
        
    Returns
    -------
    str
        The path relative to the environment variable configuration.
    """
    if env_dir := os.getenv(_ENV_VAR):
        return os.sep.join([env_dir, path])
    return None

def _site_config(path : str) -> str | None:
    """Returns the path relative to the site configuration.
    
    Parameters
    ----------
    path : str
        The relativized path.
        
    Returns
    -------
    str
        The path relative to the site configuration.
    """
    return os.sep.join([sys.prefix, _CONFIG_DIR, path]) if sys.prefix != sys.base_prefix else None

def _user_config(path : str) -> str:
    """Returns the path relative to the user configuration.
    
    Parameters
    ----------
    path : str
        The relativized path.
        
    Returns
    -------
    str
        The path relative to the user configuration.
    """
    return user_config_dir(os.sep.join([_CONFIG_DIR, path]), appauthor = False, roaming = True)

def _global_config(path : str) -> str:
    """Returns the path relative to the global configuration.
    
    Parameters
    ----------
    path : str
        The relativized path.
        
    Returns
    -------
    str
        The path relative to the global configuration.
    """
    return site_config_dir(os.sep.join([_CONFIG_DIR, path]), appauthor = False, multipath = True)

class JammiesProjectConfig:
    """Configurations within the 'project' table."""

    def __init__(self, display_warning_message: bool = True) -> None:
        """
        Parameters
        ----------
        display_warning_message : bool (default 'True')
            When True, shows a warning message when attempting to download a project file.
        """
        self.display_warning_message: bool = display_warning_message

    def list_vals(self) -> List[str]:
        """Lists all configuration options.
        
        Returns
        -------
        list[str]
            A list of all config options.
        """
        return [
            'display_warning_message'
        ]

    def encode_toml(self) -> Table:
        """Encodes the 'project' config into a table.

        Returns
        -------
        Table
            The encoded 'project' config.
        """
        project: Table = table()
        project.comment('Project related settings')
        project.add('display_warning_message',
            boolean(str(self.display_warning_message).casefold()).comment(
                'When true, shows a warning message when attempting to download a project file'
            )
        )
        return project

    @classmethod
    def decode_toml(cls, obj: DictObject) -> 'JammiesProjectConfig':
        """Decodes the 'project' table.

        Parameters
        ----------
        obj : dict[str, any]
            The encoded 'project' table.
        
        Returns
        -------
        JammiesProjectConfig
            The decoded 'project' table.
        """
        return JammiesProjectConfig(**obj)

class JammiesInternalConfig:
    """Configurations within the 'internal' table."""

    def __init__(self, generated: List[str] | None = None) -> None:
        """
        Parameters
        ----------
        generated : list[str] (default `None`)
            A list of generated files.
        """
        self.generated: Set[str] = set(generated) if generated is not None else set()

    def clear_generated_files(self) -> None:
        """Clears all generated files from the config."""
        self.generated.clear()

    def add_generated_file(self, path: str) -> None:
        """Adds a generated file to the list of generated files.

        Parameters
        ----------
        path : str
            The path of the generated file.
        """
        self.generated.add(path)

    def encode_toml(self) -> Table:
        """Encodes the 'internal' config into a table.

        Returns
        -------
        Table
            The encoded 'internal' config.
        """
        internal: Table = table()
        internal.comment('Internal data for current project (DO NOT MODIFY)')

        generated: Array = array()
        generated.comment('Files generated by the project')
        for file in self.generated:
            generated.add_line(file)
        internal.append('generated', generated)

        return internal

    @classmethod
    def decode_toml(cls, obj: DictObject) -> 'JammiesInternalConfig':
        """Decodes the 'internal' table.

        Parameters
        ----------
        obj : dict[str, any]
            The encoded 'internal' table.
        
        Returns
        -------
        JammiesInternalConfig
            The decoded 'internal' table.
        """
        return JammiesInternalConfig(**obj)

class JammiesConfig:
    """Configurations for jammies."""

    def __init__(self, project: JammiesProjectConfig = JammiesProjectConfig(),
            internal: JammiesInternalConfig = JammiesInternalConfig(),
            dirpath: str = os.curdir) -> None:
        """
        Parameters
        ----------
        project : JammiesProjectConfig (default 'JammiesProjectConfig()')
            The 'project' table within the configuration.
        dirpath : str (default '.')
            The root directory of the project.
        """
        self.project: JammiesProjectConfig = project
        self.internal: JammiesInternalConfig = internal
        self.dirpath: str = dirpath

    def list_vals(self) -> Tuple[bool, str]:
        """Lists all configuration options.
        
        Returns
        -------
        list[str]
            A list of all config options.
        """

        output: List[str] = []
        output += map(lambda s: f'project.{s}', self.project.list_vals())
        return output

    def get_val(self, name: str) -> Tuple[bool, str]:
        """Gets the value associated with the config name.
        
        Parameters
        ----------
        name : str
            The key associated with the config value.
        
        Returns
        -------
        (bool, str)
            A tuple containing whether the operation was successful
            and the associated message.
        """
        # Do not allow access to internal params
        if 'internal' in name:
            return (False, 'Accessing internal options is not allowed.')

        val: Any = self
        for key in name.split('.'):
            if not hasattr(val, key):
                return (False, f'\'{name}\' is not a valid config option.')
            val = getattr(val, key)
        return (True, str(val))

    def set_val(self, name: str, new_value: Any) -> Tuple[bool, str]:
        """Sets the value for the associated config name.
        
        Parameters
        ----------
        name : str
            The key associated with the config value.
        
        Returns
        -------
        (bool, str)
            A tuple containing whether the operation was successful
            and the associated message.
        """
        # Do not allow access to internal params
        if 'internal' in name:
            return (False, 'Accessing internal options is not allowed.')

        name_path: List[str] = name.split('.')
        # Get second to last attribute
        val: Any = self
        for key in name_path[:-1]:
            if not hasattr(val, key):
                return (False, f'\'{name}\' is not a valid config option.')
            val = getattr(val, key)

        if not hasattr(val, final_name := name_path[-1]):
            return (False, f'\'{name}\' is not a valid config option.')

        # Store previous value for update and cast type
        prev: Any = getattr(val, final_name)
        setattr(val, final_name,
            new_value := (str(new_value).casefold() == 'True'.casefold()
                if isinstance(prev, bool)
                else type(prev)(new_value))
        )

        return (True, f'{str(prev)} -> {str(new_value)}')

    def encode_toml(self, write_internal: bool = False) -> TOMLDocument:
        """Encodes the configuration.

        Returns
        -------
        TOMLDocument
            The encoded configuration.
        """
        doc: TOMLDocument = document()
        doc.add(comment('The configuration file for jammies'))
        doc.add('project', self.project.encode_toml())

        if write_internal:
            doc.add('internal', self.internal.encode_toml())
        return doc

    @classmethod
    def decode_toml(cls, obj: DictObject) -> 'JammiesConfig':
        """Decodes the configuration.

        Parameters
        ----------
        obj : dict[str, any]
            The encoded configuration.
        
        Returns
        -------
        JammiesConfig
            The decoded configuration.
        """
        return JammiesConfig(project = JammiesProjectConfig.decode_toml(
                obj['project'] if 'project' in obj else {}
            ), dirpath = obj['dirpath']
        )

    def write_config(self, scope: int = 0) -> None:
        """Writes the configuration to a file within the specified scope.

        Parameters
        ----------
        dirpath : str
            The directory of the loaded project.
        scope : int (default '0')
            A number [0, 3] representing the project, site, user, or global config, respectively.
        """
        output_path: str = config_loc(dirpath = self.dirpath, scope = scope)

        # Create directories that are missing
        os.makedirs(os.path.dirname(output_path), exist_ok = True)

        # Write config to file
        with open(output_path, mode = 'w', encoding = 'UTF-8') as file:
            dump(self.encode_toml(write_internal = scope == 0), file)

    def update_and_write(self, setter: Callable[['JammiesConfig'], Any],
            save: bool = False) -> None:
        """Updates and writes the value to the project configuration.
        
        Parameter
        ---------
        setter : (JammiesConfig) -> None
            A consumer which sets a configuration property.
        save : bool (default 'False')
            When 'True', save the configuration to the project scope.
        """
        setter(self)
        if save:
            self.write_config()

    def load_dynamic_scripts(self, logger: Logger, registrar: JammiesRegistrar) -> int:
        """Loads all scripts within the 'scripts' directories defined by the configuration.
        All scripts with the appropriate 'setup' method will be ran.

        Parameters
        ----------
        logger : Logger
            A logger for reporting on information.
        registrar : `JammiesRegistrar`
            The registrar used to register the components for the project.
        
        Returns
        -------
        int
            The number of loaded scripts.
        """

        def import_scripts(script_dir: str) -> int:
            """Imports all scripts from the current directory and executes the 'setup'
            method if present. No recursion is done.
            
            Parameters
            ----------
            script_dir : str
                The directory to look for scripts
                
            Returns
            -------
            int
                The number of loaded scripts.
            """

            scripts: int = 0

            for script in iglob(os.sep.join([script_dir, '**.py'])):
                module: ModuleType = dynamic_import(
                    _SCRIPT_DIR,
                    os.path.basename(script).rsplit(os.extsep, maxsplit = 1)[0],
                    script
                )
                if (setup_method := getattr(module, 'setup', None)) is not None:
                    logger.debug(f'Registering script: {script}')
                    setup_method(registrar)
                    scripts += 1

            return scripts

        loaded_scripts: int = 0

        # Check config locations for 'scripts' directory
        if (script_dir := _project_config(self.dirpath, _SCRIPT_DIR)) and os.path.isdir(script_dir):
            logger.debug(f'Script directory detected: {script_dir}')
            loaded_scripts += import_scripts(script_dir)
        if (script_dir := _env_var_config(_SCRIPT_DIR)) and os.path.isdir(script_dir):
            logger.debug(f'Script directory detected: {script_dir}')
            loaded_scripts += import_scripts(script_dir)
        if (script_dir := _site_config(_SCRIPT_DIR)) and os.path.isdir(script_dir):
            logger.debug(f'Script directory detected: {script_dir}')
            loaded_scripts += import_scripts(script_dir)
        if (script_dir := _user_config(_SCRIPT_DIR)) and os.path.isdir(script_dir):
            logger.debug(f'Script directory detected: {script_dir}')
            loaded_scripts += import_scripts(script_dir)
        if (script_dir := _global_config(_SCRIPT_DIR)) and os.path.isdir(script_dir):
            logger.debug(f'Script directory detected: {script_dir}')
            loaded_scripts += import_scripts(script_dir)

        return loaded_scripts

    # TODO: Deprecate for removal
    def load_dynamic_method(self, module_type: str,
            module_method: str) -> Callable[..., bool] | None:
        """Loads a script relative to the module type's directory in the
        project configurations. If the file is present, the module is dynamically
        loaded and provides a corresponding method, which returns true on successful
        operation. If the module cannot be found, then nothing will be returned.

        Parameters
        ----------
        module_type : str
            The type of module being dynamically imported.
        module_method : str
            The name of the module and method to import, formatted as '<module>:<method>'.

        Returns
        -------
        (...) -> bool
            The method loaded by the dynamic importer.
        """
        # Separate module from method and get relative path
        module, method = tuple(module_method.split(':'))
        rel_path: str = f'{os.sep.join([module_type] + module.split("."))}.py'
        module_path: str | None = None

        # Find module path
        if module == 'internal':
            return getattr(load_module(f'jammies.{module}.{module_type}'), method)

        if (abs_path := _project_config(self.dirpath, rel_path)) and os.path.exists(abs_path):
            module_path = abs_path
        elif (abs_path := _env_var_config(rel_path)) and os.path.exists(abs_path):
            module_path = abs_path
        elif (abs_path := _site_config(rel_path)) and os.path.exists(abs_path):
            module_path = abs_path
        elif (abs_path := _user_config(rel_path)) and os.path.exists(abs_path):
            module_path = abs_path
        elif (abs_path := _global_config(rel_path)) and os.path.exists(abs_path):
            module_path = abs_path

        # Load module if present
        if module_path:
            return getattr(dynamic_import(module_type, module, module_path), method)

        # Return empty processor
        print(f'Could not find module {module_method}; return empty, failing processor.')
        return None


def _update_dict(original: DictObject, merging: DictObject) -> DictObject:
    """Merges the second dictionary into the first without replacing any
    keys.

    Parameters
    ----------
    original : dict[str, any]
        The original dictionary to merge into.
    merging : dict[str, any]
        The dictionary being merged.

    Returns
    -------
    dict[str, any]
        The merged dictionary.
    """
    for key, value in merging.items():
        # Check if key isn't already present
        if key not in original:
             # Check if value is a dictionary
            if isinstance(value, dict):
                # If so, add key and update dict
                original[key] = {}
                _update_dict(original[key], merging[key])
            else:
                # Otherwise, merge key
                original[key] = value
    return original

def _read_and_update_dict(original: DictObject, path: str | None) -> DictObject:
    """Loads a dictionary, if present, and merges it into the existing
    dictionary.
    
    Parameters
    ----------
    original : dict[str, any]
        The current dictionary to merge into.
    path : str
        The path of the dictionary being merged.
    
    Returns
    -------
    dict[str, any]
        The merged dictionary.
    """
    if path and os.path.exists(path):
        with open(path, mode = 'r', encoding = 'UTF-8') as file:
            original = _update_dict(original, load(file))
    return original

def load_config(dirpath: str = os.curdir) -> JammiesConfig:
    """Loads the configuration for the project. Any settings that are not
    overridden by the project gets merged from the environment variable,
    virtual environment if present, user, and finally the global scope.

    Parameters
    ----------
    dirpath : str
        The root directory of the current project.
    
    Returns
    -------
    JammiesConfig
        The loaded configuration.
    """
    # Get project config
    config: DictObject = _read_and_update_dict({}, _project_config(dirpath, f'.{_CONFIG_FILE}'))

    # Get environment variable config
    config = _read_and_update_dict(config, _env_var_config(_CONFIG_FILE))

    # Get site config if available
    config = _read_and_update_dict(config, _site_config(_CONFIG_FILE))

    # Read user config
    config = _read_and_update_dict(config, _user_config(_CONFIG_FILE))

    # Read global config
    config = _read_and_update_dict(config, _global_config(_CONFIG_FILE))

    # Set current project directory
    config['dirpath'] = dirpath
    return JammiesConfig.decode_toml(config)

def config_loc(dirpath: str = os.curdir, scope: int = 0) -> str:
    """Gets the location of the configuration in the specified scope.

    Parameters
    ----------
    dirpath : str
        The directory of the loaded project.
    scope : int (default '0')
        A number [0, 3] representing the project, site, user, or global config, respectively.
    
    Returns
    -------
    str
        The location of the configuration.
    """
    output_path: str | None = None

    match scope:
        case 0:
            # Project config
            output_path: str = _project_config(dirpath, f'.{_CONFIG_FILE}')
        case 1:
            # Env var config if present
            if env_var := _env_var_config(_CONFIG_FILE):
                output_path: str = env_var
            # Otherwise site config if present
            elif site_var := _site_config(_CONFIG_FILE):
                output_path: str = site_var
            # Otherwise user config
            else:
                output_path: str = _user_config(_CONFIG_FILE)
        case 2:
            # User config
            output_path: str = _user_config(_CONFIG_FILE)
        case 3:
            # Global config
            output_path: str = _global_config(_CONFIG_FILE)
        case _:
            raise ValueError(f'Scope {scope} not supported, must be [0,3].')

    return output_path



"""TODO: Document, implement, finish"""

from typing import TypeAlias, Callable, TypeVar, List, Tuple
from jammies.registrar import JammiesRegistrar
from jammies.defn.file import ProjectFile, PostProcessor
from jammies.struct.codec import DictObject
from jammies.utils import input_with_default, input_yn_default

ProjectFileBuilder: TypeAlias = Callable[[JammiesRegistrar], ProjectFile]
"""A supplier used to construct a ProjectFile from an user's input."""

PF = TypeVar('PF', bound = ProjectFile)
"""The type of the project file."""

def build_file(registrar: JammiesRegistrar, codec_name: str, callback: Callable[[DictObject], PF]) -> PF:
    """Builds a ProjectFile from user input based on the
    passed in callback.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    codec_name : str
        The name of the codec for the project file.
    callback : (kwargs) -> ProjectFile
        A function that takes in the arguments of the ProjectFile
        and returns the completed ProjectFile.
    
    Returns
    -------
    ProjectFile
        The built project file.
    """
    kwargs: DictObject = {
        'codec': registrar.get_project_file_codec(codec_name),
        'name': input_with_default(ProjectFile, 'name', 'Name of the project file'),
        'rel_dir': input_with_default(ProjectFile, 'rel_dir', 'Directory to extract to'),
    }

    # Post processors
    post_processor: str = input_with_default(ProjectFile, 'post_processor',
        'Post processor on file')
    kwargs['post_processor'] = post_processor if post_processor is None \
        else (registrar.get_post_processor(post_processor), { 'type': post_processor})

    return callback(kwargs)



"""A script containing the post processor which converts a notebook
to a python file.
"""

import os
from glob import iglob
import subprocess as sp
from jammies.module import has_module
from jammies.log import Logger
from jammies.registrar import JammiesRegistrar

_REGISTRY_NAME: str = 'convert_notebook'
"""The registry name of the post processor."""

def setup(registrar: JammiesRegistrar) -> None:
    """A setup method used to register components to the project.
    
    Parameters
    ----------
    registrar : `JammiesRegistrar`
        The registrar used to register the components for the project.
    """

    # Check to see if nbconvert exists
    if has_module('nbconvert'):
        registrar.register_post_processor(_REGISTRY_NAME, notebook_to_script)
    else:
        registrar.add_post_processor_missing_message(_REGISTRY_NAME,
            'The nbconvert Python library is not installed. You can install'
            + ' the library by running of the following commands:'
            + os.linesep
            + '-> pip install nbconvert'
        )

def notebook_to_script(logger: Logger, current_dir: str) -> bool:
    """Converts a Python notebook to a script via `nbconvert`.
    
    Parameters
    ----------
    logger : Logger
        A logger for reporting on information.
    current_dir : str
        The current directory to execute the processor within.
    
    Returns
    -------
    bool
        `True` if the operation is successful.
    """
    result: bool = False
    failed: list[str] = []

    logger.debug(f'Converting notebooks in {current_dir}')

    for filename in iglob(os.sep.join([current_dir, '**']), recursive = True):
        if filename.endswith('.ipynb'):
            logger.debug(f'Converting: {filename}')
            if sp.run(
                ['jupyter', 'nbconvert', '--to', 'script', filename],
                check = False,
                stdout = sp.DEVNULL
            ).returncode == 0:
                result = True
            else:
                failed.append(filename)

    if failed:
        return logger.error(
            'Failed to convert the following scripts: '
            + os.linesep.join(list(map(lambda s: f'-> {s}', failed)))
        )

    return result
